#!/bin/bash

########### SUDO COMMANDS WITHOUT PASSWORD #############################################
# 
# Specific commands must be executable without sudo
# 
# Add to sudoers file (sudo visudo):
# 
# Disable password for specific commands
# pi ALL=(ALL:ALL) NOPASSWD: /usr/bin/chmod -R 755 /data/public/capserver/images
# pi ALL=(ALL:ALL) NOPASSWD: /usr/bin/mv /home/pi/classifier.mlmodel /data/public/capserver/
# pi ALL=(ALL:ALL) NOPASSWD: /usr/bin/mv /home/pi/classifier.version /data/public/capserver/
# pi ALL=(ALL:ALL) NOPASSWD: /usr/bin/chown -R www-data\:www-data /data/public/capserver/
#
########################################################################################

########### EXPLANATION OF RSYNC FLAGS #################################################
# 
# -h human readable output
# -v verbose output
# -r recursive
# -P print information about long-running transfers, keep partial files (—-progress --partial)
# -t preserves modification times
# -u only update if newer
# 
########################################################################################

########### PATHS ######################################################################

WORK_DIR="${HOME}/Projects/Caps/Caps-Server/Training"
BACKUP_DIR="./backup"
IMAGE_DIR="../Public/images"
VERSION_FILE="../Public/classifier.version"
MODEL_FILE="../Public/classifier.mlmodel"

TRAINING_ITERATIONS="17"
SSH_PORT="5432"
SERVER="pi@christophhagen.de"
SERVER_ROOT_PATH="/data/servers/caps/Public"

########################################################################################

echo "[INFO] Working in directory ${WORK_DIR}"
cd $WORK_DIR

retVal=$?
if [ $retVal -ne 0 ]; then
    echo '[ERROR] Directory not found'
    return $retVal
fi

echo "[INFO] Ensuring permissions for images on server..."
ssh -p $SSH_PORT ${SERVER} "sudo chmod -R 755 ${SERVER_ROOT_PATH}/images"

retVal=$?
if [ $retVal -ne 0 ]; then
    echo '[ERROR] Failed to change image permissions'
    return $retVal
fi

echo "[INFO] Transferring images from server..."
rsync -hrut --info=progress2 -e "ssh -p ${SSH_PORT}" ${SERVER}:/${SERVER_ROOT_PATH}/images/ "${IMAGE_DIR}"

retVal=$?
if [ $retVal -ne 0 ]; then
    echo '[ERROR] Failed to transfer images from server'
    return $retVal
fi

echo "[INFO] Getting classifier version from server..."
scp -P $SSH_PORT ${SERVER}:/${SERVER_ROOT_PATH}/classifier.version $VERSION_FILE

retVal=$?
if [ $retVal -ne 0 ]; then
    echo '[ERROR] Failed to get classifier version'
    return $retVal
fi

# Read classifier version from file
OLD_VERSION=$(< $VERSION_FILE)
NEW_VERSION=$(($OLD_VERSION + 1))

echo "[INFO] Backing up model ${OLD_VERSION}..."
mv $MODEL_FILE "${BACKUP_DIR}/classifier${OLD_VERSION}.mlmodel"

retVal=$?
if [ $retVal -ne 0 ]; then
    echo '[WARNING] Failed to back up old model'
fi

echo "[INFO] Training model ${NEW_VERSION} ..."
swift train.swift $IMAGE_DIR $TRAINING_ITERATIONS $MODEL_FILE

retVal=$?
if [ $retVal -ne 0 ]; then
    echo '[ERROR] Failed to train model'
    return $retVal
fi

echo "[INFO] Incrementing version file..."
echo "${NEW_VERSION}" > $VERSION_FILE

echo "[INFO] Copying the files to the server..."
scp -P $SSH_PORT $MODEL_FILE $VERSION_FILE ${SERVER}:~/

retVal=$?
if [ $retVal -ne 0 ]; then
    echo '[ERROR] Failed to copy new files to server'
    return $retVal
fi

echo "[INFO] Moving server files into public directory..."
ssh -p ${SSH_PORT} ${SERVER} "sudo mv /home/pi/classifier.* ${SERVER_ROOT_PATH}"

retVal=$?
if [ $retVal -ne 0 ]; then
    echo '[ERROR] Failed to move files on server'
    return $retVal
fi

echo "[INFO] Updating server permissions..."
ssh -p ${SSH_PORT} ${SERVER} "sudo chown -R www-data\:www-data ${SERVER_ROOT_PATH}"

retVal=$?
if [ $retVal -ne 0 ]; then
    echo '[ERROR] Failed to update file permissions on server'
    return $retVal
fi

echo "[INFO] Process finished"
